/* Entity Hook functions - Copyright (C) 2003 
 * Ravenous Bugblatter Beast <ravenousbugblatterbeast@hotmail.com>
 *
 * Incorporating APIHIJACK by Matt Pietrek & Wade Brainerd
 *
 *
 * This file solves the issue of Metamod having to export the entity functions
 * for all the entities in the game DLL. Instead it patches the import table for
 * kernel32.dll in Half-Life's swds.dll so that it any calls it makes to GetProcAddress
 * are redirected to mmGetProcAddress. mmGetProcAddress knows that if a function is not
 * exported by metamod, then it should look for it in the Game DLL instead.
 *
 *
 */


#include <extdll.h>
#include <meta_api.h>

/*--------------------------------------------------------------------------------------------------------
  APIHIJACK - Based on DelayLoadProfileDLL.CPP, by Matt Pietrek for MSJ February 2000.
  http://msdn.microsoft.com/library/periodic/period00/hood0200.htm
  Adapted by Wade Brainerd, wadeb@wadeb.com

  Simplified for this specific purpose and made cross-compiler compatible by Ravenous Bugblatter Beast
--------------------------------------------------------------------------------------------------------*/

// Macro for convenient pointer addition.
// Essentially treats the last two parameters as DWORDs.  The first
// parameter is used to typecast the result to the appropriate pointer type.
#define MakePtr(cast, ptr, addValue ) (cast)( (DWORD)(ptr)+(DWORD)(addValue))

// Default Hook Stub Structure: Contains data about the original function, Name/Ordinal, Address
// and a Count field.  This is actually a block of assembly code.




/* Compilation note:
 *
 * If you are using gcc rather than msvc to compile this, pragma pack is only enabled
 * in version 2.96 and later by default. In 2.95 you need to #define HANDLE_PRAGMA_PACK_PUSH_POP
 * in the gcc sources when you compile gcc */

#pragma pack( push, 1 )
struct DLPD_IAT_STUB
{
	BYTE    instr_CALL;
	DWORD   data_call;
	BYTE    instr_JMP;
	DWORD   data_JMP;
	DWORD   count;
	DWORD   pszNameOrOrdinal;

	DLPD_IAT_STUB() : instr_CALL( 0xE8 ), instr_JMP( 0xE9 ), count( 0 ) {}
};
#pragma pack( pop )


struct SFunctionHook
{
	char* Name;         // Function name, e.g. "DirectDrawCreateEx".
	void* HookFn;       // Address of your function.
	void* OrigFn;       // Stored by HookAPICalls, the address of the original function.
};

struct SDLLHook
{
	// Name of the DLL
	char* Name;

	// Function hook array.  Terminated with a NULL Name field.
	// NOTE: This was variable-sized in the original code but Intel C++
	// compliler barfed and MSVC said it was a non-standard extension
	SFunctionHook Functions[2];
};


//===========================================================================
// Given an HMODULE, returns a pointer to the PE header

PIMAGE_NT_HEADERS PEHeaderFromHModule(HMODULE hModule)
{
	PIMAGE_NT_HEADERS pNTHeader = 0;

	try {
		if ( PIMAGE_DOS_HEADER(hModule)->e_magic != IMAGE_DOS_SIGNATURE ) return 0;
		pNTHeader = PIMAGE_NT_HEADERS(PBYTE(hModule) + PIMAGE_DOS_HEADER(hModule)->e_lfanew);
		if ( pNTHeader->Signature != IMAGE_NT_SIGNATURE ) pNTHeader = 0;
	}
	catch(...) {
	}

	return pNTHeader;
}

//===========================================================================
// Builds stubs for and redirects the IAT for one DLL (pImportDesc)

bool RedirectIAT( SDLLHook* DLLHook, PIMAGE_IMPORT_DESCRIPTOR pImportDesc, PVOID pBaseLoadAddr )
{
	PIMAGE_THUNK_DATA pIAT;     // Ptr to import address table
	PIMAGE_THUNK_DATA pINT;     // Ptr to import names table
	PIMAGE_THUNK_DATA pIteratingIAT;

	// Figure out which OS platform we're on
	OSVERSIONINFO osvi;
	osvi.dwOSVersionInfoSize = sizeof(osvi);
	GetVersionEx( &osvi );

	// If no import names table, we can't redirect this, so bail
	if ( pImportDesc->OriginalFirstThunk == 0 ) return false;
	pIAT = MakePtr( PIMAGE_THUNK_DATA, pBaseLoadAddr, pImportDesc->FirstThunk );
	pINT = MakePtr( PIMAGE_THUNK_DATA, pBaseLoadAddr, pImportDesc->OriginalFirstThunk );

	// Count how many entries there are in this IAT.  Array is 0 terminated
	pIteratingIAT = pIAT;
	unsigned cFuncs = 0;
	while ( pIteratingIAT->u1.Function ) {
		cFuncs++;
		pIteratingIAT++;
	}

	if ( cFuncs == 0 ) return false;

	// These next few lines ensure that we'll be able to modify the IAT,
	// which is often in a read-only section in the EXE.
	DWORD flOldProtect, flNewProtect, flDontCare;
	MEMORY_BASIC_INFORMATION mbi;

	// Get the current protection attributes
	VirtualQuery( pIAT, &mbi, sizeof(mbi) );

	// remove ReadOnly and ExecuteRead attributes, add on ReadWrite flag
	flNewProtect = mbi.Protect;
	flNewProtect &= ~(PAGE_READONLY | PAGE_EXECUTE_READ);
	flNewProtect |= (PAGE_READWRITE);

	if ( !VirtualProtect(pIAT, sizeof(PVOID) * cFuncs, flNewProtect, &flOldProtect) ) return false;



	// Scan through the IAT, completing the stubs and redirecting the IAT
	// entries to point to the stubs
	pIteratingIAT = pIAT;

	while ( pIteratingIAT->u1.Function ) {
		void* HookFn = 0;  // Set to the SFunctionHook
		if ( !IMAGE_SNAP_BY_ORDINAL( pINT->u1.Ordinal ) ) { // import by name
			PIMAGE_IMPORT_BY_NAME pImportName = MakePtr( PIMAGE_IMPORT_BY_NAME, pBaseLoadAddr, pINT->u1.AddressOfData );

			// Iterate through the hook functions, searching for this import.
			SFunctionHook* FHook = DLLHook->Functions;
			while ( FHook->Name ) {
				if ( lstrcmpi( FHook->Name, (char*)pImportName->Name ) == 0 ) {
					// Save the old function in the SFunctionHook structure and get the new one.
					FHook->OrigFn = (void*)pIteratingIAT->u1.Function;
					HookFn = FHook->HookFn;
					break;
				}
				FHook++;
			}
		}


		// Replace the IAT function pointer if we have a hook.
		if ( HookFn ) {
			// Cheez-o hack to see if what we're importing is code or data.
			// If it's code, we shouldn't be able to write to it
			if ( IsBadWritePtr( (PVOID)pIteratingIAT->u1.Function, 1 ) ) {
				pIteratingIAT->u1.Function = (DWORD)HookFn;
			}
			else if ( osvi.dwPlatformId == VER_PLATFORM_WIN32_WINDOWS ) {
				// Special hack for Win9X, which builds stubs for imported
				// functions in system DLLs (Loaded above 2GB).  These stubs are
				// writeable, so we have to explicitly check for this case
				if ( (DWORD)pIteratingIAT->u1.Function > 0x80000000 ) {
					pIteratingIAT->u1.Function = (DWORD)HookFn;
				}
			}
		}

		pIteratingIAT++;    // Advance to next IAT entry
		pINT++;             // Advance to next INT entry
	}

	// Put the page attributes back the way they were.
	VirtualProtect( pIAT, sizeof(PVOID) * cFuncs, flOldProtect, &flDontCare);

	return true;
}

//===========================================================================
// Top level routine to find the EXE's imports, and redirect them
bool HookAPICalls( SDLLHook* Hook, const char* ModuleName )
{
	if ( !Hook )return false;

	HMODULE hModule = GetModuleHandle( ModuleName );
	PIMAGE_NT_HEADERS pModuleNTHdr = PEHeaderFromHModule( hModule );

	if ( !pModuleNTHdr ) return false;

	DWORD importRVA = pModuleNTHdr->OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_IMPORT].VirtualAddress;
	if ( !importRVA ) return false;

	// Convert imports RVA to a usable pointer
	PIMAGE_IMPORT_DESCRIPTOR pImportDesc = MakePtr( PIMAGE_IMPORT_DESCRIPTOR,hModule, importRVA );

	// Iterate through each import descriptor, and redirect if appropriate
	while ( pImportDesc->FirstThunk ) {
		PSTR pszImportModuleName = MakePtr( PSTR, hModule, pImportDesc->Name);
		if ( lstrcmpi( pszImportModuleName, Hook->Name ) == 0 ) {
			RedirectIAT( Hook, pImportDesc, (PVOID)hModule );
		}
		pImportDesc++;  // Advance to next import descriptor
	}

	return true;
}



/* ================================================================================================= */
/* End of code from apijack                                                                          */
/* ================================================================================================= */


/* Declarations */
FARPROC WINAPI mmGetProcAddress(HMODULE hModule,LPCSTR lpProcName);
typedef FARPROC (WINAPI *GetProcAddress_Type)(HMODULE hModule,LPCSTR lpProcName);
HINSTANCE g_hMetamod;
HINSTANCE g_hGame;


/* Our hooked function list */
SDLLHook KernelHook =
{
	"KERNEL32.DLL",
	{
		{ "GetProcAddress", mmGetProcAddress },
		{ NULL, NULL }
	}
};



/* Called from DllMain when metamod loads to install the hook */
void InitEntityHook( HINSTANCE hMetamod, HINSTANCE hGame)
{
	if (hGame==0) {
		::MessageBox(0,"Unable to locate game dll - you must use +localinfo mm_gamedll on your hlds command line",0,0);
		exit(0);
	}

    g_hMetamod=hMetamod;
	g_hGame=hGame;

	if (HookAPICalls( &KernelHook, "swds.dll" )) return;
	
	/* Failed to hook kernel32, try kernel instead */
	strcpy(KernelHook.Functions[0].Name,"KERNEL.DLL");
	
	if (HookAPICalls( &KernelHook, "swds.dll" )) return;

	::MessageBox(0,"Unable to hook GetProcAddress in the kernel - wrong operating system?",0,0);
	exit(0);
}

/* Called from DllMain when metamod unloads to uninstall the hook */
void ShutdownEntityHook()
{
	KernelHook.Functions[0].HookFn = KernelHook.Functions[0].OrigFn;
	HookAPICalls( &KernelHook, "swds.dll" );
}


/* The actual hook - all the rest of this file was just to
 * make these few lines run! */
FARPROC WINAPI mmGetProcAddress(HMODULE hModule,LPCSTR lpProcName) {

	/* First, run a normal GetProcAddress */
	FARPROC pResult=(*(GetProcAddress_Type)(KernelHook.Functions[0].OrigFn))(hModule,lpProcName);
	if (pResult) {
		return  pResult;
	}

	/* If that fails, are we looking for a Metamod function that doesn't exist?
	 * If not, give up now
	 */
	if (hModule!=g_hMetamod) return 0;

	/* Now run a normal GetProcAddress on the game DLL */
	pResult=(*(GetProcAddress_Type)(KernelHook.Functions[0].OrigFn))(g_hGame,lpProcName);
	return pResult;	
}

